package uk.co.vault101.ritual;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.audio.Sound;
import com.badlogic.gdx.graphics.Camera;
import com.badlogic.gdx.graphics.g3d.utils.CameraInputController;
import com.badlogic.gdx.math.Vector3;

public class GameCamInputController extends CameraInputController {

	private final Vector3 tmpV1 = new Vector3();
	private final Vector3 tmpV2 = new Vector3();
	public String debug; // FIXME : REMOVE THIS. A controller shouldn't have this debug junk in it
	private float scrollTotal = 0;
	private final GraphicalComponent gc;
	private final Vector3 prevTarget = new Vector3(14,0,14); // TODO initial sync these with game state, not define twice
	private final Vector3 prevCam = new Vector3(16, 6, 16);// TODO initial sync these with game state, not define twice
	private final Sound soundBump;
	private long lastPlayedSound = 0;
	
	public GameCamInputController(Camera camera, final GraphicalComponent gc) {
		super(camera);
		this.gc = gc;
		soundBump = Gdx.audio.newSound(Gdx.files.internal("sound/duff.ogg"));
	}
	
	@Override
	public boolean scrolled (int amount) {
		if (scrollTotal + amount < -5.0f) {
			return false;
		}
		
		if (scrollTotal + amount > 7.0f) {
			return false;
		}
		
		scrollTotal += amount;
		return zoom(amount * scrollFactor * translateUnits);
	}
	
	@Override
	protected boolean process(float deltaX, float deltaY, int button) {
		if (button == rotateButton) {
			debug = "deltaX: "+deltaX+" deltaY: "+deltaY;
			tmpV1.set(camera.direction).crs(camera.up).y = 0f;
			// TODO some kind of proper limits
			camera.rotateAround(target, tmpV1.nor(), deltaY * rotateAngle);
			camera.rotateAround(target, Vector3.Y, deltaX * -rotateAngle);
		} else if (button == translateButton) {
					
			camera.translate(tmpV1.set(camera.direction).crs(camera.up).nor().scl(-deltaX * translateUnits));

			tmpV2.set(camera.direction).crs(camera.up).nor().scl(-deltaY * translateUnits);
			float x = tmpV2.x;
			tmpV2.x = tmpV2.z;
			tmpV2.z = -x;
			camera.translate(tmpV2);
			
			if (translateTarget) {
				target.add(tmpV1).add(tmpV2);
			}
			
			{ // scope
				final float t = -0.5f;
				if (target.x < t) {
					camera.translate(t-target.x, 0, 0);
					target.add(t-target.x, 0, 0);
				}
				if (target.y < 0) {
					camera.translate(0, -target.y, 0);
					target.add(0, -target.y, 0);
				}
				if (target.z < t) {
					camera.translate(0, 0, t-target.z);
					target.add(0, 0, t-target.z);
				}
			}

			{ // scope
				final float t = WorldMap.MAX_WORLD_DIMENSION-0.5f;
				if (target.x > t) {
					camera.translate(t - target.x, 0, 0);
					target.add(t - target.x, 0, 0);
				}
				if (target.y > 0) {
					camera.translate(0, -target.y, 0);
					target.add(0, -target.y, 0);
				}
				if (target.z > t) {
					camera.translate(0, 0, t - target.z);
					target.add(0, 0, t - target.z);
				}
			}
			
			if (gc.map != null) { // scope
				final Tile targetTile = gc.map.getTile((int)(target.x+0.5f), (int)(target.z+0.5f));
				
				if (targetTile.isImpassable()) {
					// now move out
					if (System.currentTimeMillis() - lastPlayedSound > 2000) {
						lastPlayedSound = System.currentTimeMillis();
						soundBump.play();
					}
					target.set(prevTarget);
					camera.position.set(prevCam);
				}
			}
			prevTarget.x = target.x;
			prevTarget.y = target.y;
			prevTarget.z = target.z;
			
			prevCam.x = camera.position.x;
			prevCam.y = camera.position.y;
			prevCam.z = camera.position.z;
			
		} else if (button == forwardButton) {
			camera.translate(tmpV1.set(camera.direction).crs(camera.up).nor().scl(-deltaX * translateUnits));

			tmpV2.set(camera.direction).crs(camera.up).nor().scl(-deltaY * translateUnits);
			float x = tmpV2.x;
			tmpV2.x = tmpV2.z;
			tmpV2.z = -x;
			camera.translate(tmpV2);

			if (forwardTarget) {
				target.add(tmpV1).add(tmpV2);
			}
		}
		if (autoUpdate) {
			camera.update();
		}
		return true;
	}

	@Override
	public void update() {
		
		if (rotateRightPressed || rotateLeftPressed || forwardPressed || backwardPressed) {
			final float delta = Gdx.graphics.getDeltaTime();

			if (rotateLeftPressed) { // strafe
				tmpV1.set(camera.direction).crs(camera.up).nor().scl(delta * translateUnits);
				camera.position.add(tmpV1);
				target.add(tmpV1);
			}
			if (rotateRightPressed) { // strafe
				tmpV1.set(camera.direction).crs(camera.up).nor().scl(-delta * translateUnits);
				camera.position.add(tmpV1);
				target.add(tmpV1);
			}

			if (forwardPressed) {
				tmpV2.set(camera.direction).crs(camera.up).nor().scl(-delta * translateUnits);
				float x = tmpV2.x;
				tmpV2.x = -tmpV2.z;
				tmpV2.z = x;
				camera.translate(tmpV2);
				if (forwardTarget) {
					target.add(tmpV2);
				}
			}
			if (backwardPressed) {
				tmpV2.set(camera.direction).crs(camera.up).nor().scl(-delta * translateUnits);
				float x = tmpV2.x;
				tmpV2.x = tmpV2.z;
				tmpV2.z = -x;
				camera.translate(tmpV2);
				if (forwardTarget) {
					target.add(tmpV2);
				}
			}
			if (autoUpdate) {
				camera.update();
			}
		}
	}
}
