package uk.co.vault101.ritual;

import static com.badlogic.gdx.scenes.scene2d.actions.Actions.fadeIn;
import static uk.co.vault101.ritual.Ritual.HEIGHT;
import static uk.co.vault101.ritual.Ritual.WIDTH;

import java.nio.FloatBuffer;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.Screen;
import com.badlogic.gdx.assets.AssetManager;
import com.badlogic.gdx.graphics.Camera;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.GL20;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.Texture.TextureFilter;
import com.badlogic.gdx.graphics.g2d.BitmapFont;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.graphics.g3d.Environment;
import com.badlogic.gdx.graphics.g3d.Material;
import com.badlogic.gdx.graphics.g3d.Model;
import com.badlogic.gdx.graphics.g3d.ModelBatch;
import com.badlogic.gdx.graphics.g3d.attributes.BlendingAttribute;
import com.badlogic.gdx.graphics.g3d.attributes.ColorAttribute;
import com.badlogic.gdx.graphics.g3d.attributes.TextureAttribute;
import com.badlogic.gdx.graphics.g3d.environment.DirectionalLight;
import com.badlogic.gdx.graphics.g3d.loader.ObjLoader.ObjLoaderParameters;
import com.badlogic.gdx.math.Rectangle;
import com.badlogic.gdx.math.Vector3;
import com.badlogic.gdx.scenes.scene2d.Stage;
import com.badlogic.gdx.scenes.scene2d.ui.Label;
import com.badlogic.gdx.utils.Array;
import com.badlogic.gdx.utils.BufferUtils;
import com.badlogic.gdx.utils.viewport.StretchViewport;

import uk.co.vault101.ritual.model.GameItem;

public class GameGraphicalComponent extends GraphicalComponent {

	private GameState gs;
	
	private final Vector3 tmpVector = new Vector3();

	private final long BORN = System.currentTimeMillis();
    private final String filename = "maps/ritual.json";
    
	/** used in the class they are relevant. If they need to be used elsewhere, either reference GameState.<CONSTANTNAME> or move elsewhere */ 
	// Use fbx-conv for production use (.g3db compressed binaries)
    // instead of .obj models.

	private final String MODEL_FLATTILE = "data/flattile.obj";
	private final String MODEL_FLATTILE_2 = "data/flattile_2.obj";
	private final String MODEL_PENTAGRAM = "data/tile_pentagram.obj";
	private final String MODEL_WALL = "data/tile_wall_straight.obj";
	private final String MODEL_TARGET = "data/target.obj";

	private final String MODEL_TESTTILE = "data/testtile.obj";
	private final String MODEL_TILESELECTOR = "data/wall.obj";
	//private final String MODEL_GHOUL = "data/"
    // TODO: these are different objects where they could just be one object with a different material
    private final String MODEL_GREENSELECTOR = "data/greenselector.obj";
    private final String MODEL_REDSELECTOR = "data/redselector.obj";
    private final String MODEL_BLUESELECTOR = "data/blueselector.obj";
    private final String MODEL_PODIUM = "data/tile_podium.obj";

	// GAME ITEMS
	private final String SATANS_BED_PAN = "data/satans_bed_pan.obj";
	private final String SATANS_CHEST_HAIR = "data/satans_chest_hair.obj";
	private final String SISTER_BERTHA = "data/sister_bertha.obj";
	private final String SKULL = "data/skull.obj";
	private final String GOATS_TOE = "data/goats_toe.obj";

    private final String[] ALL_MODELS = new String[] {
    	MODEL_FLATTILE, MODEL_PENTAGRAM, MODEL_FLATTILE_2, MODEL_WALL, MODEL_TARGET, SATANS_BED_PAN, MODEL_TESTTILE, MODEL_TILESELECTOR, MODEL_GREENSELECTOR, MODEL_REDSELECTOR, MODEL_BLUESELECTOR, SATANS_CHEST_HAIR, SISTER_BERTHA, SKULL, GOATS_TOE, MODEL_PODIUM
    };
    
	/** StringBuilder used in label writing */
	private final StringBuilder sb;
	
	public final ModelBatch modelBatch;
	
	private final Environment plainEnvironment;

	private final ColorAttribute ambientLight;
	public static ColorAttribute fogColor;
	private static ColorAttribute fogRedColor;
    
	public boolean loading;
		
    /** references to models that can be 'picked' */
    private final Set<Model> pickables;
	
    /** Default expected instances */
	public final Array<GameObject> defaultInstances = new Array<>();
	
    private final BlendingAttribute blendingDefault = new BlendingAttribute();
	
    private Model modelVoid;
    private Model modelDefaultTile;
    private Model modelWall;
    private Model modelPentagram;
    private Model modelPodium;
	public GameObject target;

	/** Pentagram coverage pentagram coverage (3x3) TODO THIS IS DEFINED IN TWO PLACES! GAME JAM TIME CONSTRAINTS EXCUSE HERE! */
	public Rectangle rect;
	
    private Material selectionMaterial;

    private float oscPerSecondSin, rotationPerSecond;
	private int visibleCount;
	
	private final Label labelFPS;
	private final Label labelTile;

	private Label labelInventoryTitle;
	private Label labelInventoryItem1;
	private Label labelInventoryItem2;

	private Label labelCountDown;

	private Label labelGameOver;
	private Label labelWin;

	private BitmapFont gameFont = new BitmapFont(Gdx.files.internal("fonts/vinque32.fnt"), Gdx.files.internal("fonts/vinque32.png"), false);
	
	public GameGraphicalComponent(final Ritual display) {
		super(display);
		this.stage = new Stage();
		stage.setViewport(new StretchViewport(WIDTH,HEIGHT));
		font = new BitmapFont();
		glyphLayout.setText(font, "ABC");
		
		pickables = new HashSet<>();
		
		sb = new StringBuilder();
		
		labelFPS = new Label(" ", new Label.LabelStyle(font, Color.WHITE));
		labelTile = new Label(" ", new Label.LabelStyle(font, Color.GREEN));

		final float labelX = 5;
		labelFPS.setPosition(labelX, HEIGHT-(1*font.getLineHeight()));
		labelFPS.setVisible(false);
		
		labelTile.setPosition(labelX, HEIGHT-(2*font.getLineHeight()));
		labelTile.setVisible(false);

		createInventoryListingPlaceHolder();
		createCountdownPlaceHolder();
		createLabelGameOver();
		createLabelWin();

		stage.addActor(labelFPS);
		stage.addActor(labelTile);

		// model test
        modelBatch = new ModelBatch();
        plainEnvironment = new Environment();
        ambientLight = new ColorAttribute(ColorAttribute.AmbientLight, 0.4f, 0.4f, 0.4f, 1f);
       
        plainEnvironment.set(ambientLight);
        fogColor = new ColorAttribute(ColorAttribute.Fog, 0.0f, 0.0f, 0.0f, 1f);
        fogRedColor = new ColorAttribute(ColorAttribute.Fog, 0.5f, 0.0f, 0.0f, 1f);
        plainEnvironment.set(fogColor);        
        plainEnvironment.add(new DirectionalLight().set(1.0f, 1.0f, 1.0f, -1f, -0.8f, -0.2f));

        assets = new AssetManager();
        final ObjLoaderParameters objParams = new ObjLoaderParameters();
        objParams.flipV = true;
        objParams.textureParameter.genMipMaps = true;
        objParams.textureParameter.minFilter = TextureFilter.MipMapLinearNearest;
        objParams.textureParameter.magFilter = TextureFilter.Nearest;
        
		for (final String model : ALL_MODELS) {
			assets.load(model, Model.class, objParams);
		}

	}

	private void createLabelWin() {
		labelWin = new Label("You Win!", new Label.LabelStyle(gameFont, Color.GREEN));

		labelWin.setFontScale(5);
		labelWin.setPosition(WIDTH / 2 - 500, HEIGHT / 2 - labelGameOver.getHeight() / 2);
		labelWin.setVisible(false);

		stage.addActor(labelWin);
	}

	private void createLabelGameOver() {
		labelGameOver = new Label("GAME OVER!", new Label.LabelStyle(gameFont, Color.RED));

		labelGameOver.setFontScale(5);
		labelGameOver.setPosition(WIDTH / 2 - 700, HEIGHT / 2 - labelGameOver.getHeight() / 2);
		labelGameOver.setVisible(false);

		stage.addActor(labelGameOver);
	}

	private void createCountdownPlaceHolder() {
		labelCountDown = new Label("00:00", new Label.LabelStyle(gameFont, Color.YELLOW));

		labelCountDown.setFontScale(2);
		labelCountDown.setPosition(WIDTH / 2 - 150, labelCountDown.getHeight());
		labelCountDown.setVisible(true);

		stage.addActor(labelCountDown);
	}

	private void createInventoryListingPlaceHolder() {

		float inventoryListingLeftEdge = WIDTH - 650; // TODO warning about magic numbers like this, especially on screen width things
		float inventoryListingVerticalPosition = HEIGHT-(2* gameFont.getLineHeight());

		labelInventoryTitle = new Label("Inventory Items:-", new Label.LabelStyle(gameFont, Color.WHITE));
		labelInventoryTitle.setPosition(inventoryListingLeftEdge, inventoryListingVerticalPosition);
		labelInventoryTitle.setVisible(true);

		inventoryListingLeftEdge += 35;
		inventoryListingVerticalPosition += -40;

		labelInventoryItem1 = new Label("", new Label.LabelStyle(gameFont, Color.WHITE));
		labelInventoryItem1.setPosition(inventoryListingLeftEdge, inventoryListingVerticalPosition);
		labelInventoryItem1.setVisible(true);

		inventoryListingVerticalPosition += -40;

		labelInventoryItem2 = new Label("", new Label.LabelStyle(gameFont, Color.WHITE));
		labelInventoryItem2.setPosition(inventoryListingLeftEdge, inventoryListingVerticalPosition);
		labelInventoryItem2.setVisible(true);

		stage.addActor(labelInventoryTitle);
		stage.addActor(labelInventoryItem1);
		stage.addActor(labelInventoryItem2);
	}

	public void loadMap(final String filename) {
		// First, clean up the old one if present
		if (map != null) {
			map.dispose();
		}
		
		instances.clear();
		instances.addAll(defaultInstances);
		
		// Secondly, load the new map
		map = new WorldMap();
		
		// pentagram coverage (3x3)
		rect = new Rectangle((map.width/2)-1, (map.height/2)-1, 2, 2);
		
		final int cp = map.width/2;
		final Pair[] pairs = new Pair[] {
			new Pair(cp, cp+2),
			new Pair(cp-2, cp+1),
			new Pair(cp-2, cp-2),
			new Pair(cp+2, cp-2),
			new Pair(cp+2, cp+1)
		};
		
        if (map.generated) {
        	mapInstanceMaker(pairs);
        }
        
		if (podiums == null) {			
			podiums = new Tile[pairs.length];
			for (int i=0; i < podiums.length; i++) {
				podiums[i] = map.getTile(pairs[i].x, pairs[i].z);
			}
		}
			
        // IF we want random npcs
//		for (int i=0; i < 10; i++) {
//			final Tile t = map.determineSpawnTile();
//			map.addNPC(assets, "horseface"+System.currentTimeMillis(), t.getX(), t.getY(), t.getZ(), gs);
//		}
        
		instances.addAll(map.getInstances());
		
		map.initialiseGameItems(	  getSatansBedpanGameItem()
									, getSatansChestHairItem()
									, getSisterBerthaRosaryItem()
									, getSkullItem()
									, getGoatsToe());

		for (final GameItem gameItem : map.getAvailableGameItems()) {
			instances.add(gameItem.getGameObject());
		}
	}

	private GameItem getSatansBedpanGameItem() {
		final String name = "Satan's Bedpan";
		return new GameItem(name, new GameObject(assets.get(SATANS_BED_PAN, Model.class), name), true);
	}

	private GameItem getSatansChestHairItem() {
		final String name = "Satan's Chest Hair";
		return new GameItem(name, new GameObject(assets.get(SATANS_CHEST_HAIR, Model.class), name), true);
	}

	private GameItem getSisterBerthaRosaryItem() {
		final String name = "Sister Bertha's Rosary";
		return new GameItem(name, new GameObject(assets.get(SISTER_BERTHA, Model.class), name), true);
	}

	private GameItem getSkullItem() {
		final String name = "Skull";
		return new GameItem(name, new GameObject(assets.get(SKULL, Model.class), name), true);
	}

	private GameItem getGoatsToe() {
		final String name = "Goat's Toe";
		return new GameItem(name, new GameObject(assets.get(GOATS_TOE, Model.class), name), true);
	}

	private final Random random = new Random();
	private final int[] rotations = new int[]{0,90,180,270};
	
	public GameObject getWallObject(final Tile tile) {
		final GameObject instance = new GameObject(modelWall, MODEL_WALL);
		tile.setGameObject(instance);
		
		tile.setTranslate(tile.getX(), tile.getY(), tile.getZ());
		instance.transform.rotate(0, 1, 0, rotations[random.nextInt(rotations.length)]);
		instances.add(instance);
		tile.setPickable(true);
		tile.setImpassable(true);
		return instance;
	}
    
	private void mapInstanceMaker(final Pair[] podiumCoords) {

		int roomSize = WorldMap.ROOM_SIZE; // value of 0 creates no walls!!

		for (int z = 0; z < map.height; z++) {
			for (int x = 0; x < map.width; x++) {
				if (map.getTile(x, z) == null) {
					//continue; // 'the void'
					// TODO: or draw a 'void' tile for picking (placing)
					final Tile tile = new Tile(x, 0, z, map);
					map.setTile(x, z, tile);

					GameObject instance = null;

					if (x == 0 || z == 0 || x == map.width-1 || z == map.height-1) {
						instance = getWallObject(tile);

					} else if ((x % (roomSize+1) == 0
							&& ((z - (roomSize+1)/2) % (roomSize+1)) != 0)
						|| (z % (roomSize+1) == 0
							&&  ((x - (roomSize+1)/2) % (roomSize+1)) != 0)) {

						instance = getWallObject(tile);
					} else {
						instance = getDefaultGameObject(x, z, podiumCoords);

					}
					tile.setGameObject(instance);

					tile.setTranslate(x, 0, z);
					if (instance.gon.equals(MODEL_PENTAGRAM)) { // allow random rotations on all but the pentagram
						instance.transform.rotate(0, 1, 0, 270);
					} else {
						instance.transform.rotate(0, 1, 0, rotations[random.nextInt(rotations.length)]);
					}
					instances.add(instance);
					tile.setPickable(true);
				}
				// TODO read existing Tile objects

			}
		}
	}

    private GameObject getDefaultGameObject(int x, int z, final Pair[] podiumCoords) {
    	   	
    	if (x == map.width/2 && z == map.height/2) {
    		// center, pentagram
    		return new GameObject(modelPentagram, MODEL_PENTAGRAM);
    	} else if (rect.contains(x, z)) {
    		return new GameObject(modelVoid, "das void");
    	} else {
    		final Tile here = map.getTile(x, z);

			for (final Pair p : podiumCoords) {
				if (p.x == here.getX() && p.z == here.getZ()) {
					return new GameObject(modelPodium, "podium joe");
				}
			}

    		return new GameObject(modelDefaultTile, MODEL_TESTTILE);
    	}
    }

	@Override
	public void setScreenRef(Screen screen) {
		gs = (GameState) screen;
	}
	
	@Override
	protected void update3D(final float delta) {
		rotationPerSecond += delta;
		
		oscPerSecondSin = (((float)Math.sin((System.currentTimeMillis()-BORN)/200f)) / 2f) + 0.5f;
		
		if (map != null) {
			map.update(delta);
	        for (final GameItem gi : map.getAvailableGameItems()) {
	        	if (gi.getGameObject() != null) {        	
	        		blendingDefault.opacity = 0.25f + oscPerSecondSin;
	        		gi.getGameObject().materials.first().set(blendingDefault);
//	        		gi.getGameObject().transform.setToTranslation(gi.getGameObject().tileRef.getX(), gi.getGameObject().tileRef.getY(), gi.getGameObject().tileRef.getZ());
//	        		gi.getGameObject().transform.rotate(0, 1, 0, (0.5f * rotationPerSecond * 360) % 360);
	        	}
	        }
		}
		
		// Additional (controller) effect on camera
		final float axisMultiplier = 0.01f;
		// translation with LEFT stick
		gs.getCamController().process(axisMultiplier * XBox360Pad.axisLeftXValue, axisMultiplier * XBox360Pad.axisLeftYValue, gs.getCamController().translateButton);
		// rotation with LEFT stick
		gs.getCamController().process(axisMultiplier * XBox360Pad.axisRightXValue, axisMultiplier * XBox360Pad.axisRightYValue, gs.getCamController().rotateButton);
		// zoom with TRIGGERS LT RT
		if (gs.getCam().position.y > 2f &&  -XBox360Pad.axisTriggerValue > 0) {
			gs.getCamController().zoom(-XBox360Pad.axisTriggerValue);
		}
		if (gs.getCam().position.y < 9f &&  -XBox360Pad.axisTriggerValue < 0) {
			gs.getCamController().zoom(-XBox360Pad.axisTriggerValue);
		}
				
        gs.getCamController().update();
        if (target != null) {
        	target.transform.setToTranslation(gs.getCamController().target);
	        // rotation speed * rotations per second  
        	target.transform.rotate(0, 1, 0, (0.5f * rotationPerSecond * 360) % 360);
        }
//        
//        if (gs.getTileSelector() != null) {        	
//        	blendingDefault.opacity = oscPerSecondSin;
//        	gs.getTileSelector().materials.first().set(blendingDefault);
//        }        
	}

	@Override
	protected void render3D(final float delta) {
		 
	    Gdx.gl.glViewport(0, 0, Gdx.graphics.getWidth(), Gdx.graphics.getHeight());

	    if (gs != null && gs.timeExpired) {
	        plainEnvironment.set(fogRedColor);
	    	Gdx.gl.glClearColor(0.5f, 0.0f, 0.0f, 1f);
	    } else {
	    	Gdx.gl.glClearColor(0.0f, 0.0f, 0.0f, 1f);
	    }
        Gdx.gl.glClear(GL20.GL_COLOR_BUFFER_BIT | GL20.GL_DEPTH_BUFFER_BIT);

        // here is where you're do stage.act(delta) and stage.draw if you were mental 
        // and wanted a background and UI elements and everything UNDER or BEHIND your 3D scene
        // Generally, don't do this!
        
        modelBatch.begin(gs.getCam());
        visibleCount = 0;
        
        for (final GameObject instance : instances) {
        	//instance.setTranslate(instance.x, instance.y+(float)Math.sin(System.currentTimeMillis()/10), instance.z);
            // TODO : if you want to see frustum freeze, set visibility on the GameObject instance
        	// then the frustum can be 'frozen' to see what the world looks like from another angle.
          if (isVisible(gs.getCam(), instance)) {
      		// This block below demonstrates that blocks could be made to go up and down, (sine wave) if desired
        	    if (instance.tileRef != null && instance.tileRef.getGameItem() != null) {
        	    	float arg = 0.7f*((((float)Math.sin((System.currentTimeMillis()-BORN)/1000f)) / 4f) + 0.5f);
        	    	instance.tileRef.getGameItem().getGameObject().transform.setToTranslation(instance.tileRef.getX(), arg, instance.tileRef.getZ());
        	    }
                modelBatch.render(instance, plainEnvironment);
                visibleCount++;
          }
        }
        
        modelBatch.end();
	}

	/**
	 * Check the GameObject against the camera frustum to see if it's visible in the scene or not
	 * 
	 * @param cam Camera to test against
	 * @param instance GameObject to test
	 * @return boolean (true/false)
	 */
    protected boolean isVisible(final Camera cam, final GameObject instance) {
    	// TODO - this does center -> radius checking which is not ideal for tiles as the extent (note: diameter) is root(2) = 1.4142135623730950488016887242097
    	instance.transform.getTranslation(tmpVector);
        tmpVector.add(instance.center);
        return cam.frustum.sphereInFrustum(tmpVector, instance.radius);
    }
    
	@Override
	protected void update2D(final float delta) {
		
        if (loading && assets.update()) {
            doneLoading();
        }
        
        labelInventoryTitle.setColor(gs != null && gs.cannaPickThatUp ? Color.RED : Color.WHITE);
        labelInventoryItem1.setColor(gs != null && gs.cannaPickThatUp ? Color.ORANGE : Color.WHITE);
        labelInventoryItem2.setColor(gs != null && gs.cannaPickThatUp ? Color.ORANGE : Color.WHITE);
        
        // FPS counter and other debug-only useful stats
        updateFPSStats();
        updateInventoryItems();
		updateCountDownTimer();

		showWinLabel();

        stage.act(delta);
	}

	private void showWinLabel() {
		if (gs.gameWon) {
			labelWin.setVisible(true);
		}
	}
	private void updateCountDownTimer() {

		if (!gs.timeExpired && gs.countdownRunning) {
			float deltaTime = Gdx.graphics.getDeltaTime(); //You might prefer getRawDeltaTime()

			gs.timeRemaining -= deltaTime; //if counting down

			int minutes = ((int) gs.timeRemaining) / 60;
			int seconds = ((int) gs.timeRemaining) % 60;


			if (gs.timeRemaining <= 0) {
				gs.timeExpired = true;
			}
			labelCountDown.setText(String.format("%02d:%02d", minutes, seconds));
		}
		
		if (gs.timeExpired) {
			labelCountDown.setColor(Color.RED);
		}
	}

	private void updateInventoryItems() {

		labelInventoryItem1.setText("1: " + ((gs.getCollectedItems().size() > 0) ? gs.getCollectedItems().get(0).getName() : ""));

		labelInventoryItem2.setText("2: " + ((gs.getCollectedItems().size() > 1) ?  gs.getCollectedItems().get(1).getName() : ""));
	}

	/**
	 * Method for updating FPS counter and other statistics
	 * and the labels on screen for showing this data.
	 */
	private void updateFPSStats() {
		sb.setLength(0);
		
		sb.append("FPS: ").append(Gdx.graphics.getFramesPerSecond());
		
		sb.append(" Tex: "+gs.texturePointer);
		
		sb.append(" CAM (" + gs.getCam().position.x + "," + gs.getCam().position.y + "," + gs.getCam().position.z + ")");
		
		if (target != null) {
			sb.append(" TGT (" +gs.getCamController().target.x + "," + gs.getCamController().target.y + "," + gs.getCamController().target.z + ")");
		}
		
		sb.append(" Vis: "+visibleCount);
		sb.append(" tempI: "+tempInstances.size);
		
		final long memused = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
		sb.append(" Mem: "+(memused/(1024*1024))+" M");
		labelFPS.setText(sb);
		
		if (gs.getSelectedTile() != null) {
			labelTile.setText(""+gs.getSelectedTile().toJSON());
		}
	}
	
	
	private void doneLoading() {

        initPreviewMaterials();

        modelVoid = new Model();
        modelDefaultTile = assets.get(MODEL_TESTTILE, Model.class);
        modelWall = assets.get(MODEL_WALL, Model.class);
        modelPentagram = assets.get(MODEL_PENTAGRAM, Model.class);
        modelPodium = assets.get(MODEL_PODIUM, Model.class);
        
        // anything that's a 'floor' is pickable
        pickables.add(modelVoid);
        pickables.add(modelDefaultTile);
        pickables.add(modelWall);
        pickables.add(modelPentagram);
        pickables.add(modelPodium);

        loadAnyQueuedAssets();
        
        //  MUST define defaultInstances first BEFORE loading map
        defaultInstances.add(initTarget());
        //defaultInstances.add(initTileSelector());

        loadMap(filename);

        loadAnyQueuedAssets();
        
        gs.selectTileScreenCenter();

		// Add inventory text
        
		gs.countdownRunning = true;

        loading = false;

    }
	
	private void initPreviewMaterials() {
        // load texture once
        final Texture tex = new Texture(Gdx.files.internal("data/tileset.png"), true);
        tex.setFilter(TextureFilter.MipMapLinearNearest, TextureFilter.Nearest);
        
        // Anisotropic filtering
        if (true) {
	        tex.bind();
	        final FloatBuffer buffer = BufferUtils.newFloatBuffer(64);
	        Gdx.gl20.glGetFloatv(GL20.GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, buffer);
	        Gdx.gl20.glTexParameterf(GL20.GL_TEXTURE_2D, GL20.GL_TEXTURE_MAX_ANISOTROPY_EXT, buffer.get(0));
        }
        
        previewMaterial = new Material[64];
        for (int i=0; i < previewMaterial.length; i++) {
        	previewMaterial[i] = new Material();
        	final TextureAttribute ta = new TextureAttribute(TextureAttribute.Diffuse);
        	final float u = ((float)(i%8))/8.0f;                    // X Position Of Current Character
        	final float v = ((float)(i/8))/8.0f;                    // Y Position Of Current Character
        	final float u2 = u+(1.0f/8.0f);
        	final float v2 = v+(1.0f/8.0f);
        	ta.set(new TextureRegion(tex, u, v, u2, v2));
        	previewMaterial[i].set(ta);
        }
	}
	
	/**
	 * Blocking - will sit here and load until queue is empty
	 */
	private void loadAnyQueuedAssets() {
		while (assets.getQueuedAssets() > 0) {
			assets.update();
		}
	}

	private GameObject initTarget() {
    	// 'target' item 
    	final Model targetModel = assets.get(MODEL_TARGET, Model.class);
    	target = new GameObject(targetModel, MODEL_TARGET);
    	target.transform.setToTranslation(gs.getCamController().target);
    	return target;
	}

	private GameObject initTileSelector() {
        selectionMaterial = new Material();
        blendingDefault.opacity = 0.4f;
        selectionMaterial.set(ColorAttribute.createDiffuse(1.0f, 0.61f, 0f, 0.5f), blendingDefault);
        
    	// tile selector
    	final Model ts = assets.get(MODEL_TILESELECTOR, Model.class);
    	final GameObject go = new GameObject(ts, MODEL_TILESELECTOR);
    	gs.setTileSelector(go);
    	final Material mat = gs.getTileSelector().materials.get(0);
        mat.clear();
        mat.set(selectionMaterial); // set whatever the model provided to the custom made one from this class file
        return go;
	}
	
	
	@Override
	protected void render2D(final float delta) {
    	stage.getBatch().begin();
    
    	font.setColor(Color.WHITE);
    	glyphLayout.setText(font, gs.debugString);
    	font.draw(stage.getBatch(), gs.debugString, (WIDTH/2)-(glyphLayout.width/2), glyphLayout.height*2);
    	
    	stage.getBatch().end();
    	
    	stage.draw();
	}

	@Override
	public void show() {
        loading = true;
		stage.getRoot().getColor().a = 0;
		stage.getRoot().addAction(fadeIn(0.5f));
	}

	@Override
	public void hide() {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void pause() {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void resume() {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void resize(final int width, final int height) {
    	WIDTH = width;
    	HEIGHT = height;
		stage.getViewport().update(width, height, true);
	}

	@Override
	public void dispose() {
		if (stage != null) {
			stage.dispose();
		}
		
		// model stuff
		modelBatch.dispose();
	    assets.dispose();
	    instances.clear();
	}
	
	public void nextScreen() {
		labelGameOver.setVisible(true);
		//switchScreen(display, display.getGameState());
	}
	
	private class Pair {
		public int x;
		public int z;
		public Pair (int x, int z) {
			this.x = x;
			this.z = z;
		}
		public String toString() {
			return ""+x+","+z;
		}
	}
}
