package uk.co.vault101.ritual;

import org.json.JSONException;
import org.json.JSONObject;

import com.badlogic.gdx.graphics.g3d.Model;
import com.badlogic.gdx.graphics.g3d.ModelInstance;
import com.badlogic.gdx.math.Vector3;
import com.badlogic.gdx.math.collision.BoundingBox;


/**
 * A GameObject for handling 3D models, their bounding boxes,
 * the center of the objects and dimensions
 */
public class GameObject extends ModelInstance implements ToJSON {
	
	public static long statsCreated;
    public final Vector3 center = new Vector3();
    public final Vector3 dimensions = new Vector3();
    public float radius;
    public final String gon;
    public Tile tileRef = null;
    public final BoundingBox bounds = new BoundingBox();

    public GameObject(final Model model, final String gon) {
        super(model);
        statsCreated++;
        this.gon = gon;
        recalculateBounds(); // required upon create or frustum culling goes funky
    }
    
    public void recalculateBounds() {
    	calculateBoundingBox(bounds);
        bounds.getCenter(center);
        bounds.getDimensions(dimensions);
        radius = dimensions.len() / 2f;
    }
    
    /** Used in saving the object, helps with debugging. Native/binary/compressed can wait */
    @Override
    public JSONObject toJSON() {
		final JSONObject jo = new JSONObject();
		try {
			jo.put("c", "GameObject"); // class
			jo.put("radius", radius);
			jo.put("name", gon);
			jo.put("center", center.toString());
			jo.put("dim", dimensions.toString());
			jo.put("bounds", bounds.toString());
		} catch (JSONException e) {};
		return jo;
    }
}
