package uk.co.vault101.ritual;

import com.badlogic.gdx.graphics.GL20;
import com.badlogic.gdx.utils.Timer;

import uk.co.vault101.ritual.GraphicalComponent;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.Input;
import com.badlogic.gdx.InputAdapter;
import com.badlogic.gdx.InputMultiplexer;
import com.badlogic.gdx.Screen;
import com.badlogic.gdx.audio.Music;
import com.badlogic.gdx.audio.Sound;
import com.badlogic.gdx.controllers.Controller;
import com.badlogic.gdx.controllers.ControllerListener;
import com.badlogic.gdx.controllers.Controllers;
import com.badlogic.gdx.controllers.PovDirection;
import com.badlogic.gdx.graphics.PerspectiveCamera;
import com.badlogic.gdx.graphics.g3d.Material;
import com.badlogic.gdx.math.Rectangle;
import com.badlogic.gdx.math.Vector3;
import com.badlogic.gdx.math.collision.Ray;

import uk.co.vault101.ritual.ai.NPC;
import uk.co.vault101.ritual.model.GameItem;

import java.util.ArrayList;
import java.util.List;

public class GameState extends InputAdapter implements Screen, ControllerListener {

	private final GraphicalComponent gc;

	protected long start;
	private Music music;
	private final Sound soundItemPickup;
	private final Sound soundItemDropoffGood;
	private final Sound soundRitualTinkles;
	
	public String debugString = "";

	// If multiplayer this needs to be per character in game!
	private List<GameItem> collectedItems = new ArrayList<>();

	/** Cameras. They are in GameState rather than graphical component because of their reliance on InputAdapter */
	public final PerspectiveCamera cam;
	private final GameCamInputController camController;
	private GameObject tileSelector;
	public int texturePointer = 0;

	public boolean timeExpired = false;
	public float timeRemaining = 40;
	public boolean countdownRunning = false;

	public boolean gameWon = false;
	public boolean cannaPickThatUp = false;

	// pentagram coverage (3x3) TODO THIS IS DEFINED IN TWO PLACES! GAME JAM TIME CONSTRAINTS EXCUSE HERE!
	public Rectangle rect;
	private int podiumDropoff = 0;

	// TODO abit crappy but cant think of neater way right now
	public boolean restarting = false;

	public GameState(final GraphicalComponent gc) {
		this.gc = gc;
		gc.setScreenRef(this);
		//music = Gdx.audio.newMusic(Gdx.files.internal("sound/volatile-reaction.ogg"));
		music = Gdx.audio.newMusic(Gdx.files.internal("sound/clash_defiant.ogg"));
		music.setLooping(true);
		
		soundItemPickup = Gdx.audio.newSound(Gdx.files.internal("sound/39041__wildweasel__itemup2.ogg"));
		soundItemDropoffGood = Gdx.audio.newSound(Gdx.files.internal("sound/99636-yeah.ogg"));
		//soundRitualTinkles = Gdx.audio.newSound(Gdx.files.internal("sound/77248__jobro__celestial-temple.ogg"));
		soundRitualTinkles = Gdx.audio.newSound(Gdx.files.internal("sound/darkness_speaks.ogg"));
		
        cam = new PerspectiveCamera(67, Gdx.graphics.getWidth(), Gdx.graphics.getHeight());
        cam.position.set(14, 5.5f, 14);
        cam.lookAt(16, 0, 16);
        cam.near = 0.1f;
        // FIXME cam.far WOULD be 10f default but libgdx fog coords start/end currently locked to cam.near/far which is rubbish
        cam.far = 10f;
        cam.update();

        camController = new GameCamInputController(cam, gc);
        camController.target.set(16, 0, 16);
        camController.scrollTarget = false;
        
        Controllers.addListener(this); // receives events from all controllers

        Gdx.input.setInputProcessor(new InputMultiplexer(this, camController));
	}

	private void reset() {
		music.stop();
	}

	public List<GameItem> getCollectedItems() {
		return collectedItems;
	}

	public GameObject getTileSelector() {
		return tileSelector;
	}
    
    public void setTileSelector(final GameObject ts) {
		tileSelector = ts;
	}
    
	public Tile getSelectedTile() {
		return selected;
	}
	
	public GameCamInputController getCamController() {
		return camController;
	}
	
	public PerspectiveCamera getCam() {
		return cam;
	}
	
	@Override
	public void dispose() {
		gc.dispose();
	}

	@Override
	public void hide() {
		gc.hide();
	}

	@Override
	public void pause() {
		gc.pause();
	}
	
	@Override
	public void render(final float delta) {
		gameLogic(delta);
		gc.render(delta);

		if (timeExpired) {
			loseCondition();
			gc.nextScreen();
		}
	}
	
	private void gameLogic(final float delta) {
		if (gc.map == null) {
			return;
		}
		if (rect == null) {
			rect = new Rectangle((gc.map.width/2)-1, (gc.map.height/2)-1, 2, 2);
		}
		if (!timeExpired) {
			itemPickupDropoff();
		}
	}
	
	private void itemPickupDropoff() {
		// consider stuffing this in PLayer object
		final Tile currentTile = gc.map.getTile((int)(camController.target.x+0.5f), (int)(camController.target.z+0.5f));
		cannaPickThatUp = false;
		{ // scope
			final GameItem gameItem = currentTile.getGameItem();
			if (gameItem != null && gameItem.isCollectable()) {
				if (collectedItems.size() < 2) {
					soundItemPickup.play();
					collectedItems.add(gameItem);
					gc.instances.removeValue(gameItem.getGameObject(), true);
					currentTile.clearGameItem();
				} else {
					// flash the inventory
					debugString = "Can't pick that up!";
					cannaPickThatUp = true;
				}
			}
		}
		
		// drop off zone
		if (rect.contains(currentTile.getX(), currentTile.getZ())) {
			
			if (!collectedItems.isEmpty()) {
				// TODO the difference between a good dropoff and a bad one
				soundItemDropoffGood.play();
				final GameGraphicalComponent ggc = (GameGraphicalComponent) gc;
				System.out.println("GameState: # collected Items: "+collectedItems.size());
				for (final GameItem gi : collectedItems) {
					for (final Tile t : ggc.podiums) {
						if (t == null) {
							System.out.println("GameState WARNING: podium tile is null");
							continue;
						}
						if (t.getGameItem() == null) {
							gi.setNotCollectable();
							t.setGameItem(gi);
							gc.instances.add(gi.getGameObject());
							podiumDropoff++;
							if (podiumDropoff == gc.podiums.length) {
								winCondition();
							}
							break;
						}
					}
				}
				
				// TODO drop renderable instances 
				collectedItems.clear();
			}
		}
		
		
	}
		
	private void loseCondition() {
		soundRitualTinkles.play();
		XBox360Pad.axisRightXMoving = true;
		XBox360Pad.axisRightXValue = -0.1f;
		
		XBox360Pad.axisRightYMoving = false;
		XBox360Pad.axisRightYValue = 0.0f;
		XBox360Pad.axisLeftYMoving = false;
		XBox360Pad.axisLeftYValue = 0.0f;
		XBox360Pad.axisLeftXMoving = false;
		XBox360Pad.axisLeftXValue = 0.0f;
		gameWon = false;
		countdownRunning = false;
	}
	
	private void winCondition() {
		camController.target.set(gc.map.width/2, 0, gc.map.height/2);
		XBox360Pad.axisRightXMoving = true;
		XBox360Pad.axisRightXValue = -0.1f;
		
		XBox360Pad.axisRightYMoving = false;
		XBox360Pad.axisRightYValue = 0.0f;
		XBox360Pad.axisLeftYMoving = false;
		XBox360Pad.axisLeftYValue = 0.0f;
		XBox360Pad.axisLeftXMoving = false;
		XBox360Pad.axisLeftXValue = 0.0f;
		soundRitualTinkles.play();
		gameWon = true;
		countdownRunning = false;
	}

	@Override
	public void resize(int width, int height) {
		gc.resize(width, height);
	}

	@Override
	public void resume() {
		gc.resume();
	}

	@Override
	public void show() {
		start = System.currentTimeMillis();
		gc.show();
		
		
		music.play();
	}
	
	/** As it says, this is a temporary vector, generally passed into methods to be populated */
	private final Vector3 tmpVector = new Vector3();
	
	/**
     * CAUTION: There's an assumption here that the object Y coord (3D worldspace) will always be at 0
     * If you use this to pick objects where their Y != 0, you WILL get strange results
     * @param screenX 2D X coord (screenspace)
     * @param screenY 2D Y coord (screenspace)
     * @param setTarget boolean whether or not to update the camera target we render
     * @return The pickable GameObject at the 2D screen X,Y supplied
     */
    private Tile pickTile2D (final int screenX, final int screenY, final boolean setTarget) {
    	if (gc.map == null) {
    		return null;
    	}
        final Ray ray = cam.getPickRay(screenX, screenY);
        
        final float dist = -ray.origin.y / ray.direction.y;
        tmpVector.set(ray.direction).scl(dist).add(ray.origin);
        
        if (tmpVector.x < gc.map.width+0.5f && tmpVector.z < gc.map.height+0.5f
        		&& tmpVector.x > -0.5f && tmpVector.z > -0.5f) {
        	
        	final Tile tile = gc.map.getTile((int)(tmpVector.x+0.5f), (int)(tmpVector.z+0.5f));
        	if (tile != null && tile.isPickable()) {
        		if (setTarget) {
        			camController.target.set(tmpVector.x,tmpVector.y,tmpVector.z);
        		}
        		return tile;
        	}
        }
        return null;
    }
    
    private Tile selected = null;
	private Tile selecting = null;
    
	@Override
    public boolean touchDragged(int screenX, int screenY, int pointer) {
		
//    	return false;
		return true;
    }
    
	@Override
	public boolean touchDown(final int screenX, final int screenY, final int pointer, final int button) {
//    	if (button == camController.translateButton) {
//    		selecting = pickTile2D(screenX, screenY, true);
//    		if (selecting != null) {
//    			// TODO  ?
//    		}
//    	} else if (button == camController.rotateButton) {
//    		final Tile prev = selected;
//
//    		selecting = pickTile2D(screenX, screenY, true);
//    		if (selecting != null) {
//    			setSelected(selecting);
//
//    			if (prev == selecting) {
//        			// TODO already select on same tile - what to do here?
//        		}
//    		} // else null and we don't care
//    	}
    	
    	return false;
	}
	
	/**
	 * Set the supplied GameObject to 'selected' state.
	 * @param value
	 */
    public void setSelected (final Tile value) {    
    	if (tileSelector != null && value != null) {
    		tileSelector.transform.setToTranslation(value.getX(), value.getY(), value.getZ());
    	}
    	selected = value;
    }
    
    private void deselectTiles() {
    	selecting = null;
    	selected = null;
    }
    
	@Override
    public boolean touchUp (final int screenX, final int screenY, final int pointer, final int button) {

		return true;
//		return false;
    }
	

	@Override
	public boolean keyDown(int keycode) {
		switch (keycode) {
		case Input.Keys.ESCAPE:
			pause();
			dispose();
			Gdx.app.exit();
			break;
//		case Input.Keys.I: // toggle if a tile is impassable
//			if (selecting != null) {
//				gc.map.setDirtybit();
//				selecting.setImpassable(!selecting.isImpassable());
//			}
//			break;
//		case Input.Keys.W:
//			if (selecting != null) {
//				makeWall();
//			}
//			break;
		}
		return true;
	}



	@Override
    public boolean keyTyped(char character) {
	    // consider escape = quit game
		// dispose correctly and exit
    	return false;
    }
	
	@Override
	public boolean accelerometerMoved(Controller controller, int accelerometerCode, Vector3 value) {
		if (!countdownRunning) return false;
		debugString = "Accel. "+accelerometerCode+" "+value;
		return false;
	}

	@Override
	public boolean axisMoved(Controller controller, int axisCode, float value) {
		if (!countdownRunning) return false;
		switch (axisCode) {
			case XBox360Pad.AXIS_RIGHT_X:
				if (value > 0.2 || value < -0.2) {// deadzone
					debugString = "Axis "+axisCode+" "+value;
				    XBox360Pad.axisRightXValue = value;
				} else {
					XBox360Pad.axisRightXMoving = false;
					XBox360Pad.axisRightXValue = 0.0f;
				}
				break;
			case XBox360Pad.AXIS_RIGHT_Y:
//				if (value > 0.2 || value < -0.2) {// deadzone
//					debugString = "Axis "+axisCode+" "+value;
//					XBox360Pad.axisRightYValue = value;
//				} else {
//					XBox360Pad.axisRightYMoving = false;
//					XBox360Pad.axisRightYValue = 0.0f;
//				}
				break;
			case XBox360Pad.AXIS_LEFT_X:
				if (value > 0.2 || value < -0.2) {// deadzone
					debugString = "Axis "+axisCode+" "+value;
					XBox360Pad.axisLeftXValue = -value;
				} else {
					XBox360Pad.axisLeftXMoving = false;
					XBox360Pad.axisLeftXValue = 0.0f;
				}
				break;
			case XBox360Pad.AXIS_LEFT_Y:
				if (value > 0.2 || value < -0.2) {// deadzone
					debugString = "Axis "+axisCode+" "+value;
					XBox360Pad.axisLeftYValue = value;
				} else {
					XBox360Pad.axisLeftYMoving = false;
					XBox360Pad.axisLeftYValue = 0.0f;
				}
				break;
			case XBox360Pad.AXIS_TRIGGER:
				XBox360Pad.axisTriggerValue = value;
				if (value > 0.01) {
					// left trigger
					debugString = "LT "+axisCode+" "+value;
				} else if (value < -0.01) {
					// right trigger
					debugString = "RT "+axisCode+" "+value;
				} else {
					XBox360Pad.axisTriggerMoving = false;
					XBox360Pad.axisTriggerValue = 0.0f;
				}
				break;
	        default:
	        	System.out.println("Unknown [AXISMOVED] "+controller.getName()+" Axis: "+axisCode+" Value: "+value);
		}
		
		return false;
	}

	public void selectTileScreenCenter() {
 		final Tile prev = selected;

		selecting = pickTile2D(Ritual.WIDTH/2, Ritual.HEIGHT/2, true);
		if (selecting != null) {
			setSelected(selecting);

			if (prev == selecting) {
    			// TODO second touch on same tile - what to do?
    		}
		} // else null and we don't care
	}
	
	private void decorate() {
		gc.map.setDirtybit();
		
		if (selecting !=  null) {
			// TODO : shift this to Tile class
			final Material mat = selecting.getGameObject().materials.first();
			mat.clear();
			mat.set(gc.previewMaterial[texturePointer]);
			selecting.setMaterialIndex(texturePointer);
			debugString = "Set selected tile material to "+texturePointer+" ("+selecting.toString()+")";
		} else {
			debugString = "TODO : did you mean to select this first before decorating?";
		}
	}
	
	public void mapSave() {
		if (gc.map.save()) {
			// successful
			System.out.println("Map saved: "+gc.map.filename);
		} else {
			System.out.println("Error saving map "+gc.map.filename+" (check log)");
		}
	}

	
	/**
	 * TODO toggle between making wall or testtile, as appropriate
	 */
	private void makeWall() {
		// make impassable
		if (selected != null) {
			gc.map.setDirtybit();
			selected.setImpassable(!selected.isImpassable());
			
			// change model/texture as appropriate
			// TODO casting crazy talk. Don't do this
			selected.setGameObject(((GameGraphicalComponent)gc).getWallObject(selected));
		}
	}
	
	@Override
	public boolean buttonDown(Controller controller, int buttonCode) {
		if (!countdownRunning) {

			switch (buttonCode) {
				case XBox360Pad.BUTTON_A:
				case XBox360Pad.BUTTON_B:
				case XBox360Pad.BUTTON_X:
				case XBox360Pad.BUTTON_Y:
					restartGame();
					break;
			}

			return false;
		}

		debugString = "Button Down "+buttonCode;
		
		switch (buttonCode) {
		case XBox360Pad.BUTTON_A:
			//selectTileScreenCenter();
			break;
		case XBox360Pad.BUTTON_B:
			//deselectTiles();
			break;
		case XBox360Pad.BUTTON_X:
			//makeWall();
			break;
		case XBox360Pad.BUTTON_Y:
			//decorate();
			break;
		case XBox360Pad.BUTTON_START:
			//mapSave();
			break;
		}
		
		return false;
	}


	private void restartGame() {

		// Dispose and clear current screen
//		gc.hide();
//		Gdx.graphics.getGL20().glClearColor( 1, 1, 1, 1 );
//		Gdx.graphics.getGL20().glClear(GL20.GL_COLOR_BUFFER_BIT | GL20.GL_DEPTH_BUFFER_BIT);

		reset();

		if (! restarting) { // defensive as otherwise restart was called multiple times!
			restarting = true;
			// Restart the game state?
			gc.switchScreen(gc.display, new GameState(new GameGraphicalComponent(gc.display)));

			//gc.dispose();
		}

	}

	@Override
	public boolean buttonUp(Controller controller, int buttonCode) {
		if (!countdownRunning) return false;
		debugString = "Button Up "+buttonCode;
		switch (buttonCode) {
		case XBox360Pad.BUTTON_RB:
			if (texturePointer < gc.previewMaterial.length - 1) {
				texturePointer++;
			} else {
				texturePointer = 0;
			}
			break;
		case XBox360Pad.BUTTON_LB:
			if (texturePointer > 0) {
				texturePointer--;
			} else {
				texturePointer = gc.previewMaterial.length - 1;
			}
			break;
		}
		
		return false;
	}

	@Override
	public boolean povMoved(Controller controller, int povCode, PovDirection value) {
		if (!countdownRunning) return false;
		debugString = "Pov "+povCode+" "+value;
		return false;
	}
	
	@Override
	public boolean xSliderMoved(Controller controller, int sliderCode, boolean value) {
		if (!countdownRunning) return false;
		debugString = "X slider "+sliderCode+" "+value;
		return false;
	}

	@Override
	public boolean ySliderMoved(Controller controller, int sliderCode, boolean value) {
		if (!countdownRunning) return false;
		debugString = "Y slider "+sliderCode+" "+value;
		return false;
	}
	
	@Override
	public void connected(Controller controller) {
		debugString = "Connected "+controller.getName();
	}

	@Override
	public void disconnected(Controller controller) {
		debugString = "Disconnected "+controller.getName();
	}	
}
