package uk.co.vault101.ritual.ai;

import java.util.Random;

import org.json.JSONException;
import org.json.JSONObject;

import uk.co.vault101.ritual.GameObject;
import uk.co.vault101.ritual.GameState;
import uk.co.vault101.ritual.Tile;
import uk.co.vault101.ritual.ToJSON;
import uk.co.vault101.ritual.WorldMap;
import uk.co.vault101.ritual.ai.pathfinding.Path.Step;

import com.badlogic.gdx.assets.AssetManager;
import com.badlogic.gdx.graphics.g3d.Model;
import com.badlogic.gdx.graphics.g3d.loader.ObjLoader.ObjLoaderParameters;
import com.badlogic.gdx.math.Vector3;

public class NPC extends Unit implements ToJSON {

	private final Vector3 tmpTarget = new Vector3();
	private final GameObject gameObject;
	private final String MODEL = "data/skull.obj";
	public final String name;
	private final Random random = new Random();
	private final int MIN_WAIT_TIME = 3000;
	private final int MAX_WAIT_TIME = 6000;
	private long waittime = 0;
	private GameState gs;
	
	public NPC (final AssetManager assets, final float x, final float y, final float z, final String name, final WorldMap map, final GameState gs) {
		super(new Vector3(x,y,z));
		this.name = name;
		this.gs = gs;
		allowDiagonalMovement = true;
		diagonalCost = true;
		speed = 1f;
		
		final ObjLoaderParameters objParams = new ObjLoaderParameters();
        objParams.flipV = true;

        assets.load(MODEL, Model.class, objParams);
        // TODO: assets.update until model is ACTUALLY loaded
        while (!assets.isLoaded(MODEL)) {
        	assets.update();
        }

		gameObject = new GameObject(assets.get(MODEL, Model.class), MODEL);
    	// TODO: initial translation should be read from map/changed/set as appropriate from other process like starting position
		gameObject.transform.setToTranslation(position);
	}

	/**
	 * Get the associated GameObject for this NPC
	 * @return GameObject for this NPC
	 */
	public GameObject getGameObject() {
		return gameObject;
	}
	
	@Override
	public void update(final float delta, final WorldMap map) {
		super.update(delta, map);
		waittime = 0;
		
		decideWhereToGo(map);
		
		// if our position has changed, do something about that
		if (!position.equals(positionLast)) {
			gameObject.transform.setToTranslation(position);
			gameObject.transform.rotate(0, 1, 0, angle);
			final Tile newTile = map.getTile(Math.round(position.x), Math.round(position.z)); 
			if (tile == null || !tile.equals(newTile)) {
				tile = newTile;
			}
			
			// sync the two
			positionLast.x = position.x;
			positionLast.y = position.y;
			positionLast.z = position.z;
			angleLast = angle;
		} else if (angle != angleLast) {
			// A similar 'previous' check for rotation. Only do the expensive transform if rotation has actually changed.
			gameObject.transform.rotate(0, 1, 0, angle);
			angleLast = angle;
		}
		
		if (tile != null) { // this only needed here if we're on top of a moving tile
			// TODO lerp distance between current and target
			position.y = tile.getY(); // adjust Y to current tile height
		}
	}
	
	private void decideWhereToGo(final WorldMap map) {
		if (path == null && waittime == 0) {
			
			if (Math.random() < 0.5f) {
				setTileTarget(map.determineSpawnTile());
			} else {
				setTileTarget(map.getTile((int)(gs.getCamController().target.x+0.5), (int)(gs.getCamController().target.z+0.5)));
			}
			
		}
	}

	@Override
	Vector3 pathfinding(final WorldMap map) {
		
		if (target != null && path == null) {
			// position/target to 'Tile' index
			// 6.51 -> 7.0 -> 7.49 are all '7' when Math.round is used
			final int rpx = Math.round(position.x);
			final int rpz = Math.round(position.z);
			final int rtx = Math.round(target.x);
			final int rtz = Math.round(target.z);
			if (rpx == rtx && rpz == rtz) {
				return null;
			}
			map.clearVisited();
			path = map.getAStarPathFinder(allowDiagonalMovement).findPath(this, rpx, rpz, rtx, rtz, diagonalCost);
		}
		
		if (path != null) {
			if (currentStep < path.getLength()) {
				final Step step = path.getStep(currentStep);
				tmpTarget.x = step.getX();
				tmpTarget.y = position.y;
				tmpTarget.z = step.getY();
				// TODO consider Y of the new vector should be set correctly!!! 
				return tmpTarget;
			} else {
				// reached end of path
				//Console.write(this, "At end of path. step: "+currentStep+" pos: "+position);
				path = null;
				waittime = System.currentTimeMillis() + (random.nextInt((MAX_WAIT_TIME - MIN_WAIT_TIME) + 1) + MIN_WAIT_TIME);
				return null;
			}
		} else {
			return null;
		}
	}
	
	public long getWaitTime() {
		return waittime;
	}

	/** Used in saving the object, helps with debugging. Native/binary/compressed can wait */
	@Override
	public JSONObject toJSON() {
		final JSONObject jo = new JSONObject();
		try {
			jo.put("c", "NPC") // class
			  .put("name", name)
			  .put("x", position.x)
			  .put("y", position.y)
			  .put("z", position.z)
			  .put("rot", angle)
			  .put("waittime", waittime);
			
		} catch (JSONException e) {};
		return jo;
	}
}
