package uk.co.vault101.ritual.ai;

import uk.co.vault101.ritual.Tile;
import uk.co.vault101.ritual.WorldMap;
import uk.co.vault101.ritual.ai.pathfinding.Mover;
import uk.co.vault101.ritual.ai.pathfinding.Path;

import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.math.Vector3;

public abstract class Unit implements Mover, Renderable {
	
	public final Vector3 position;
	public final Vector3 positionLast;
	public Tile tile;
	
	public static long statsCreated;
	public float angle;
	public float angleLast;
	public Vector3 target;
	public float life;
	public float speed;
	public int currentStep = 1;
	public Path path;
	/** PathFinder hint to smooth out paths. For goblins, animals etc consider keeping this false. True for predictable players, false for 'weaving' */
	public boolean diagonalCost;
	protected boolean allowDiagonalMovement;
	
	public Unit(final Vector3 position) {
		statsCreated++;
		this.position = new Vector3(position);
		this.positionLast = new Vector3(position);
	}

	/** Every 'Unit' must do some kind of pathfinding */
	abstract Vector3 pathfinding(final WorldMap map);
	
	@Override
	public void render(final float delta) {
		// Note, it's really the update() method that's important here.
		// render() will have to be overridden
		// If there's anything super-core for ALL units rendering, put it here
	}
	
	public boolean setTileTarget(final Tile tile) {
		boolean diff = true;
		if (target == null) {
			target = new Vector3();
		}
		currentStep = 1;
		path = null; // clear

		target.x = tile.getX();
		target.y = tile.getY();
		target.z = tile.getZ();
		return diff;
	}
	
	@Override
	public void update(final float delta, final WorldMap map) {
		
		final Vector3 tmpTarget = pathfinding(map);

		if (tmpTarget != null) {
			// forward vector from position to target
			final Vector3 newForwardVector = new Vector3(tmpTarget).sub(position).nor();
			
			// add to position, the normalised forward vector scaled by (speed * delta)
			position.mulAdd(newForwardVector, speed * delta);
			
            final Vector2 facing = new Vector2(newForwardVector.x, newForwardVector.z);
            angle = facing.angle(Vector2.Y);
            
            if (tmpTarget.epsilonEquals(position, 0.02f)) {
                currentStep++;
            }
        } else {
        	currentStep = 1;
        }
		
	}
}
